/**
 * PETALS - PETALS Services Platform. Copyright (c) 2006 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id: SoapComponent.java 154 25 sept. 06 alouis $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.binding.soap;

import java.io.File;
import java.util.Iterator;
import java.util.logging.Level;

import javax.jbi.JBIException;
import javax.naming.Context;
import javax.xml.namespace.QName;

import org.apache.axiom.om.OMAbstractFactory;
import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.OMFactory;
import org.apache.axis2.AxisFault;
import org.apache.axis2.Constants;
import org.apache.axis2.context.ConfigurationContext;
import org.apache.axis2.context.ConfigurationContextFactory;
import org.apache.axis2.deployment.DeploymentConstants;
import org.apache.axis2.description.AxisModule;
import org.apache.axis2.description.Flow;
import org.apache.axis2.description.Parameter;
import org.apache.axis2.description.TransportInDescription;
import org.apache.axis2.engine.AxisConfiguration;
import org.apache.axis2.phaseresolver.PhaseException;
import org.apache.axis2.phaseresolver.PhaseHolder;
import org.apache.axis2.phaseresolver.PhaseMetadata;
import org.apache.axis2.transport.jms.AxisJMSException;
import org.apache.axis2.transport.jms.JMSConstants;
import org.apache.axis2.transport.jms.JMSListener;
import org.ow2.petals.binding.soap.listener.incoming.PetalsReceiver;
import org.ow2.petals.binding.soap.listener.incoming.SoapExternalListenerManager;
import org.ow2.petals.component.framework.bc.AbstractBindingComponent;
import org.ow2.petals.component.framework.su.AbstractServiceUnitManager;

import static org.ow2.petals.binding.soap.Constants.Axis2.AXIS2_XML;
import static org.ow2.petals.binding.soap.Constants.Component.TRANSFORMER_SYSTEM_PROPERY_NAME;
import static org.ow2.petals.binding.soap.Constants.Component.TRANSFORMER_SYSTEM_PROPERY_VALUE;

/**
 * The SOAP binding component.
 * 
 * @author alouis - EBM Websourcing
 * @author chamerling - EBM Websourcing
 */
public class SoapComponent extends AbstractBindingComponent {

	/**
	 * The listener processing SOAP request providing from external clients.
	 * Incoming SOAP requests.
	 */
	protected SoapExternalListenerManager externalListenerManager;

	/**
	 * The SOAP component context
	 */
	protected SoapComponentContext soapContext;

	/**
	 * The PEtALS receiver
	 */
	private PetalsReceiver petalsReceiver;

	/**
	 * <p>
	 * Create the JMS transport layer including a default connection factory.
	 * </p>
	 * <p>
	 * The default connection factory is defined into the components extensions.
	 * See component extensions for more information about the configuration of
	 * this deafult connection factory.
	 * </p>
	 * <p>
	 * If the component extensions does not include needed values, the JMS
	 * transport layer will be activated without default JMS connection factory.
	 * </p>
	 * <p>
	 * The JMS transport layer is defined as (following the syntax of
	 * axis2.xml), italic bold value are component extensions:<br>
	 * <code>&lt;transportReceiver name="jms" class="org.apache.axis2.transport.jms.JMSListener"&gt;<br>
	 * &lt;parameter name="default" locked="false"><br>
	 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&lt;parameter name="java.naming.factory.initial" locked="false"&gt;<b><i>org.apache.activemq.jndi.ActiveMQInitialContextFactory</i></b>&lt;/parameter&gt;<br>
	 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&lt;parameter name="java.naming.provider.url" locked="false"&gt;<b><i>tcp://localhost:61616</i></b>&lt;/parameter&gt;<br>        
	 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&lt;parameter name="transport.jms.ConnectionFactoryJNDIName" locked="false"&gt;<b><i>QueueConnectionFactory</i></b>&lt;/parameter&gt;<br>
	 * &nbsp;&nbsp;&nbsp;&lt;/parameter&gt;<br>
	 * &lt;/transportReceiver&gt;</code>
	 * </p>
	 * 
	 * @param axisConfigurationContext
	 * @throws AxisFault
	 */
	private void createJMSTransporter(final ConfigurationContext axisConfigurationContext) throws AxisFault {

		final TransportInDescription jmsTransportDescription = new TransportInDescription(Constants.TRANSPORT_JMS);
		jmsTransportDescription.setReceiver(new JMSListener());
		axisConfigurationContext.getAxisConfiguration().addTransportIn(jmsTransportDescription);

		final String jndiInitialFactory = getComponentExtensions().get(org.ow2.petals.binding.soap.Constants.JmsTransportLayer.JNDI_INITIAL_FACTORY);
		final String jndiProviderUrl = getComponentExtensions().get(org.ow2.petals.binding.soap.Constants.JmsTransportLayer.JNDI_PROVIDER_URL);
		final String confacJndiName = getComponentExtensions().get(org.ow2.petals.binding.soap.Constants.JmsTransportLayer.CONFAC_JNDINAME);
		if (jndiInitialFactory != null && jndiProviderUrl != null && confacJndiName != null && jndiInitialFactory.length() > 0 && jndiProviderUrl.length() > 0
				&& confacJndiName.length() > 0) {

			soapContext.setJmsJndiInitialFactory(jndiInitialFactory);
			soapContext.setJmsJndiProviderUrl(jndiProviderUrl);
			soapContext.setJmsConnectionFactoryName(confacJndiName);

			getLogger().info("Create the default JMS connection factory ('" + JMSConstants.DEFAULT_CONFAC_NAME + "'):");
			getLogger().info("\t" + org.ow2.petals.binding.soap.Constants.JmsTransportLayer.JNDI_INITIAL_FACTORY + ": " + jndiInitialFactory);
			getLogger().info("\t" + org.ow2.petals.binding.soap.Constants.JmsTransportLayer.JNDI_PROVIDER_URL + ": " + jndiProviderUrl);
			getLogger().info("\t" + org.ow2.petals.binding.soap.Constants.JmsTransportLayer.CONFAC_JNDINAME + ": " + confacJndiName);

			final OMFactory omFactory = OMAbstractFactory.getOMFactory();

			final OMElement initialContextFactory = omFactory.createOMElement(new QName(DeploymentConstants.TAG_PARAMETER));
			initialContextFactory.addAttribute(DeploymentConstants.ATTRIBUTE_NAME, Context.INITIAL_CONTEXT_FACTORY, null);
			initialContextFactory.setText(jndiInitialFactory);

			final OMElement providerUrl = omFactory.createOMElement(new QName(DeploymentConstants.TAG_PARAMETER));
			providerUrl.addAttribute(DeploymentConstants.ATTRIBUTE_NAME, Context.PROVIDER_URL, null);
			providerUrl.setText(jndiProviderUrl);

			final OMElement connectionFactoryName = omFactory.createOMElement(new QName(DeploymentConstants.TAG_PARAMETER));
			connectionFactoryName.addAttribute(DeploymentConstants.ATTRIBUTE_NAME, JMSConstants.CONFAC_JNDI_NAME_PARAM, null);
			connectionFactoryName.setText(confacJndiName);

			final OMElement defaultConnectionFactory = omFactory.createOMElement(new QName(DeploymentConstants.TAG_PARAMETER));
			defaultConnectionFactory.addChild(initialContextFactory);
			defaultConnectionFactory.addChild(providerUrl);
			defaultConnectionFactory.addChild(connectionFactoryName);

			jmsTransportDescription.addParameter(new Parameter(JMSConstants.DEFAULT_CONFAC_NAME, defaultConnectionFactory));

		} else {
			getLogger().info("The JMS transport layer configuration is not complete. It is disabled.");
		}

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seeorg.ow2.petals.component.framework.bc.AbstractBindingComponent#
	 * createServiceUnitManager()
	 */
	@Override
	protected AbstractServiceUnitManager createServiceUnitManager() {

		return new SoapSUManager(this);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.objectweb.petals.component.common.AbstractComponent#doInit()
	 */
	@Override
	protected void doInit() throws JBIException {

		// create the SOAP component context
		soapContext = new SoapComponentContext(getContext(), getComponentConfiguration(), getLogger());

		// the axis2.xml is required
		final File axis2File = new File(getContext().getWorkspaceRoot(), AXIS2_XML);
		if (!axis2File.exists()) {
			throw new JBIException("Can not get axis2 configuration file");
		}

		// get axis configuration context
		getLogger().fine("Creating Axis configuration context...");
		ConfigurationContext axisConfigurationContext = null;
		try {
			axisConfigurationContext = ConfigurationContextFactory.createConfigurationContextFromFileSystem(getContext().getWorkspaceRoot(), axis2File
					.getAbsolutePath());

			createJMSTransporter(axisConfigurationContext);

			getLogger().fine("Axis configuration context created.");
			soapContext.setAxis2ConfigurationContext(axisConfigurationContext);
		} catch (final AxisFault e) {
			getLogger().log(Level.WARNING, e.getMessage());
			throw new JBIException("Can not initialize SOAP BC", e);
		}

		// create the receiver
		petalsReceiver = new PetalsReceiver(this, getChannel(), getLogger());

		// set SOAP HTTP Server
		externalListenerManager = new SoapExternalListenerManager(this, getChannel(), (AbstractServiceUnitManager) getServiceUnitManager(), soapContext,
				petalsReceiver, getLogger());

		// for each module, we add its flows (if any) to the corresponding axis
		// phases
		engageModulesHandlers(axisConfigurationContext.getAxisConfiguration());

		((SoapSUManager) getServiceUnitManager()).init(soapContext, getContext(), externalListenerManager.getSoapServerConfig(), petalsReceiver, getLogger());
		if (System.getProperty(TRANSFORMER_SYSTEM_PROPERY_NAME) == null) {
			System.setProperty(TRANSFORMER_SYSTEM_PROPERY_NAME, TRANSFORMER_SYSTEM_PROPERY_VALUE);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.objectweb.petals.component.common.AbstractComponent#doStart()
	 */
	@Override
	public void doStart() throws JBIException {
		try {
		    AxisConfiguration axisConfiguration = soapContext.getAxis2ConfigurationContext().getAxisConfiguration();
			externalListenerManager.start(axisConfiguration);
			
			// Start the JMS transport layer
			try {
                axisConfiguration.getTransportIn(Constants.TRANSPORT_JMS).getReceiver().start();
			} catch (final AxisJMSException e) {
				getLogger().info("The JMS Transport is not available.");
				getLogger().log(Level.FINE, "Unable to start the JMS Transport (" + e.getMessage() + "). SOAP over JMS is not available.", e);
			}
		} catch (final AxisFault e) {
			getLogger().severe(e.getMessage());
			throw new JBIException(e);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.objectweb.petals.component.common.AbstractComponent#doStop()
	 */
	@Override
	public void doStop() throws JBIException {
		try {
			externalListenerManager.stop();
			// Stop the JMS transport layer
			soapContext.getAxis2ConfigurationContext().getAxisConfiguration().getTransportIn(Constants.TRANSPORT_JMS).getReceiver().stop();
		} catch (final AxisFault e) {
			getLogger().severe(e.getMessage());
			throw new JBIException(e);
		}
	}

	/**
	 * For each Axis phase, we add the corresponding module's handlers if any.
	 * 
	 * @param module
	 *            the module to process
	 * @param axisConfiguration
	 *            the axis configuration
	 * @throws PhaseException
	 *             can the thrown by method addHandler()
	 */
	private void engageModuleHandlersIntoPhases(final AxisModule module, final AxisConfiguration axisConfiguration) throws PhaseException {
		Flow moduleFlow = null;
		PhaseHolder phaseHolder = null;

		getLogger().fine("Engaging module " + module.getName() + " in flows");

		// for each phase, we add the corresponding module's handlers if any
		for (int type = PhaseMetadata.IN_FLOW; type <= PhaseMetadata.FAULT_OUT_FLOW; type++) {

			if (type == PhaseMetadata.IN_FLOW) {

				phaseHolder = new PhaseHolder(axisConfiguration.getInFlowPhases());
				moduleFlow = module.getInFlow();

			} else if (type == PhaseMetadata.OUT_FLOW) {

				phaseHolder = new PhaseHolder(axisConfiguration.getOutFlowPhases());
				moduleFlow = module.getOutFlow();

			} else if (type == PhaseMetadata.FAULT_IN_FLOW) {

				phaseHolder = new PhaseHolder(axisConfiguration.getInFaultFlowPhases());
				moduleFlow = module.getFaultInFlow();

			} else if (type == PhaseMetadata.FAULT_OUT_FLOW) {

				phaseHolder = new PhaseHolder(axisConfiguration.getOutFaultFlowPhases());
				moduleFlow = module.getFaultOutFlow();
			}

			if (moduleFlow != null && phaseHolder != null) {
				for (int j = 0; j < moduleFlow.getHandlerCount(); j++) {
					phaseHolder.addHandler(moduleFlow.getHandler(j));
				}
			}
		}
	}

	/**
	 * @param axisConfiguration
	 * @throws JBIException
	 */
	@SuppressWarnings("unchecked")
	private void engageModulesHandlers(final AxisConfiguration axisConfiguration) throws JBIException {
		final Iterator<AxisModule> modules = axisConfiguration.getEngagedModules().iterator();
		while (modules.hasNext()) {
			try {
				engageModuleHandlersIntoPhases(modules.next(), axisConfiguration);
			} catch (final PhaseException e) {
				getLogger().log(Level.WARNING, e.getMessage());
				throw new JBIException("Can not initialize SOAP BC", e);
			}
		}
		getLogger().fine("Axis modules flows added");
	}

	/**
	 * @return
	 */
	public SoapExternalListenerManager getExternalListenerManager() {
		return externalListenerManager;
	}

	/**
	 * @return
	 */
	public SoapComponentContext getSoapContext() {
		return soapContext;
	}

}
