/**
 * PETALS - PETALS Services Platform. Copyright (c) 2005 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.binding.soap.listener.incoming.jetty;

import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.axis2.AxisFault;
import org.apache.axis2.Constants;
import org.apache.axis2.addressing.EndpointReference;
import org.apache.axis2.context.ConfigurationContext;
import org.apache.axis2.context.MessageContext;
import org.apache.axis2.context.SessionContext;
import org.apache.axis2.description.Parameter;
import org.apache.axis2.description.TransportInDescription;
import org.apache.axis2.engine.ListenerManager;
import org.apache.axis2.transport.TransportListener;
import org.mortbay.jetty.Connector;
import org.mortbay.jetty.Server;
import org.mortbay.jetty.handler.HandlerCollection;
import org.mortbay.jetty.nio.SelectChannelConnector;
import org.mortbay.jetty.servlet.Context;
import org.mortbay.jetty.servlet.ServletHolder;
import org.mortbay.log.Log;
import org.mortbay.thread.BoundedThreadPool;
import org.ow2.petals.binding.soap.SoapComponentContext;
import org.ow2.petals.binding.soap.listener.incoming.SoapServerConfig;
import org.ow2.petals.binding.soap.listener.incoming.servlet.ListServicesServlet;
import org.ow2.petals.binding.soap.listener.incoming.servlet.SoapServlet;
import org.ow2.petals.binding.soap.listener.incoming.servlet.WelcomeServlet;

/**
 * Create a Jetty server that will be the axis2 transportlistener. This Jetty
 * server is used to replace the SimpleHTTPServer provided by axis that have
 * quite bad performances.
 * 
 * @author Christophe Hamerling - EBM WebSourcing
 * 
 */
public class SoapServer implements TransportListener {

	/**
	 * The logger
	 */
	protected Logger logger;

	/**
	 * The axis2 configuration context (built from axis2.xml config file)
	 */
	protected ConfigurationContext configurationContext;

	/**
	 * The component properties
	 */
	protected SoapServerConfig config;

	/**
	 * The Jetty server
	 */
	protected Server server;

	/**
	 * The jetty thread pool. Size is configured from component configuration.
	 */
	protected BoundedThreadPool threadPool;

	protected ServerStats stats;

	public static final int HEADER_BUFFER_SIZE = 4096 * 4;

	static {
	    System.setProperty("org.mortbay.log.class", JettyNullLogger.class.getName());
	    Log.getLog();
	}

	/**
	 * Creates a new instance of {@link SoapServer}
	 * 
	 * @param configContext
	 * @param config
	 * @param componentContext
	 * @param logger
	 * @throws AxisFault
	 */
	public SoapServer(final SoapServerConfig config, final SoapComponentContext componentContext, final Logger logger) throws AxisFault {
		this.config = config;
		this.logger = logger;
		this.configurationContext = componentContext.getAxis2ConfigurationContext();

		// jetty logger
		Log.setLog(new JettyLogger(logger));
		
		// jetty threapool configuration
		threadPool = new BoundedThreadPool();
		threadPool.setName("BCSoapJettyThreadPool");
		threadPool.setMaxThreads(this.config.getJettyThreadMaxPoolSize());
		threadPool.setMinThreads(this.config.getJettyThreadMinPoolSize());

		// jetty http connector configuration
		final SelectChannelConnector nioConnector = new SelectChannelConnector();
		nioConnector.setPort(this.config.getPort());

		// If we assign the host, we will only be able to contact server
		// on it. No value or a null one is a wildcard so connection is possible
		// on network interface
		// @see java.net.InetSocketAddress
		if (this.config.isRestricted()) {
			nioConnector.setHost(this.config.getHostAddress());
		}

		nioConnector.setHeaderBufferSize(HEADER_BUFFER_SIZE);
		nioConnector.setStatsOn(false);
		nioConnector.setAcceptors(this.config.getJettyAcceptors());

		server = new Server();
		server.setConnectors(new Connector[] { nioConnector });
		server.setThreadPool(threadPool);

		stats = new ServerStats();

		// create context handlers
		//final ContextHandlerCollection contexts = new ContextHandlerCollection();
		final HandlerCollection contexts = new HandlerCollection();
		server.setHandler(contexts);

		// create the axis context
		final Context axisContext = new Context(contexts, "/" + this.config.getServicesContext(), Context.SESSIONS);
		axisContext.setErrorHandler(new PetalsErrorHandler(false));

		// create axis servlet holder
		final ServletHolder axisServlet = new ServletHolder(new SoapServlet(configurationContext, stats));
		axisServlet.setName("AxisServlet");
		axisServlet.setInitOrder(1);
		axisContext.addServlet(axisServlet, "/" + this.config.getServicesMapping() + "/*");
		axisContext.addServlet(axisServlet, "/servlet/AxisServlet");

		// add our own simple service list servlet
		final ServletHolder listServlet = new ServletHolder(new ListServicesServlet(configurationContext, this.config));
		listServlet.setName("ServicesListServlet");
		listServlet.setInitOrder(1);
		axisContext.addServlet(listServlet, "/" + this.config.getServicesMapping() + "/" + ListServicesServlet.MAPPING_NAME);
		
		// add welcome servlet
		final Context welcomeContext = new Context(contexts, "/", Context.SESSIONS);
		final ServletHolder welcomeServlet = new ServletHolder(new WelcomeServlet(this.config, stats));
		welcomeServlet.setName("WelcomeServlet");
		welcomeServlet.setInitOrder(1);
		welcomeContext.addServlet(welcomeServlet, "/*");
		
		configureListenerManager();
	}

	/**
	 * Configure the listener manager used by Axis
	 * 
	 * @throws AxisFault
	 */
	private void configureListenerManager() throws AxisFault {
		ListenerManager listenerManager = configurationContext.getListenerManager();
		final TransportInDescription trsIn = new TransportInDescription(Constants.TRANSPORT_HTTP);
		trsIn.setReceiver(this);
		if (listenerManager == null) {
			listenerManager = new ListenerManager();
			listenerManager.init(configurationContext);
		}
		listenerManager.addListener(trsIn, true);
	}

	public void destroy() {
		// TODO Auto-generated method stub
	}

	/**
	 * 
	 * @return
	 */
	public ConfigurationContext getConfigurationContext() {
		return configurationContext;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.apache.axis2.transport.TransportListener#getEPRsForService(java.lang
	 * .String, java.lang.String)
	 */
	public EndpointReference getEPRForService(final String serviceName, final String ip) throws AxisFault {
		return getEPRsForService(serviceName, ip)[0];
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.apache.axis2.transport.TransportListener#getEPRsForService(java.lang
	 * .String, java.lang.String)
	 */
	public EndpointReference[] getEPRsForService(final String serviceName, final String ip) throws AxisFault {
		if (config != null) {
			return new EndpointReference[] { new EndpointReference(config.getProtocol() + "://" + config.getHostAddress() + ":" + config.getPort() + "/"
					+ configurationContext.getServiceContextPath() + "/" + serviceName) };
		} else {
			throw new AxisFault("Unable to generate EPR for the transport : http");
		}
	}

	public SessionContext getSessionContext(final MessageContext messageContext) {
		// TODO Auto-generated method stub
		return null;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.apache.axis2.transport.TransportListener#init(org.apache.axis2.context
	 * .ConfigurationContext,
	 * org.apache.axis2.description.TransportInDescription)
	 */
	public void init(final ConfigurationContext axisConf, final TransportInDescription transprtIn) throws AxisFault {
		configurationContext = axisConf;

		// These values are defined by the component and not by the Axis2
		// configuration file so we set them if Axis2 do something with them...
		Parameter portParam = transprtIn.getParameter(PARAM_PORT);
		if (portParam == null) {
			portParam = new Parameter(PARAM_PORT, Integer.toString(config.getPort()));
		}
		transprtIn.addParameter(portParam);

		Parameter hostParam = transprtIn.getParameter(HOST_ADDRESS);
		if (hostParam == null) {
			hostParam = new Parameter(HOST_ADDRESS, config.getHostAddress());
		}
		transprtIn.addParameter(hostParam);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.apache.axis2.transport.TransportListener#start()
	 */
	public void start() throws AxisFault {
		logger.info("Starting Jetty server...");
		logger.info(config.getHostToDisplay() + " - Port : " + config.getPort()
				+ " - Jetty Max poolsize : " + config.getJettyThreadMaxPoolSize() + " - Jetty Min poolsize : " + config.getJettyThreadMinPoolSize()
				+ " - Jetty Acceptors size : " + config.getJettyAcceptors());
		try {
			stats.setStartTime(System.currentTimeMillis());
			server.start();
		} catch (final Exception e) {
			logger.log(Level.SEVERE, "Can not start the Jetty server");
			throw new AxisFault("Can not start the Jetty server", e);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.apache.axis2.transport.TransportListener#stop()
	 */
	public void stop() throws AxisFault {
		logger.log(Level.INFO, "Stopping Jetty server...");

		try {
			// stop jetty server
			server.stop();
			stats.setStopTime(System.currentTimeMillis());
		} catch (final InterruptedException e) {
			throw AxisFault.makeFault(e);
		} catch (final Exception e) {
			throw AxisFault.makeFault(e);
		}
	}
	
	/**
	 * Redirect specified URI to the specified web service address
	 * @param from The URI to redirect
	 * @param to The serviceName or address of the destination service
	 */
	public void addRedirect(String from, String to) {
		config.addRedirect(from,
				"/" + this.config.getServicesContext() + "/" + this.config.getServicesMapping()
				+ "/" + to);
	}
	public void removeRedirect(String from) {
		config.removeRedirect(from);
	}
}
